<?php

namespace BS\ExtendedSearch\Source\Formatter;

use BS\ExtendedSearch\Wildcarder;
use ConfigException;
use MediaWiki\Linker\LinkRenderer;
use MWException;

class Base {
	/**
	 * Used to separate multiple values in arrays
	 * when they are displayed in the UI
	 */
	public const VALUE_SEPARATOR = ', ';

	/**
	 * Used to indicate there are more values than
	 * can be displayed
	 */
	public const MORE_VALUES_TEXT = '...';

	public const AC_RANK_NORMAL = 'normal';
	public const AC_RANK_SECONDARY = 'secondary';
	public const AC_RANK_TOP = 'top';

	/**
	 *
	 * @var \BS\ExtendedSearch\Source\Base
	 */
	protected $source;

	/**
	 *
	 * @var \BS\ExtendedSearch\Lookup
	 */
	protected $lookup;

	/**
	 * @var LinkRenderer
	 */
	protected $linkRenderer;

	/**
	 *
	 * @param \BS\ExtendedSearch\Source\Base $source
	 */
	public function __construct( $source ) {
		$this->source = $source;
		// Just for convenience, as many of the formatters would use it
		$this->linkRenderer = $this->source->getBackend()->getService( 'LinkRenderer' );
	}

	/**
	 * Sets current instance of Lookup object that the
	 * result being formatted
	 *
	 * @param \BS\ExtendedSearch\Lookup $lookup
	 */
	public function setLookup( $lookup ) {
		$this->lookup = $lookup;
	}

	/**
	 * Convenience function - returns RequestContext object
	 *
	 * @return \RequestContext
	 */
	public function getContext() {
		return $this->source->getBackend()->getContext();
	}

	/**
	 * Returns structure of the result for each source
	 * It allows sources to modify default result structure
	 *
	 * @param array $defaultResultStructure
	 * @return array
	 */
	public function getResultStructure( $defaultResultStructure = [] ) {
		return $defaultResultStructure;
	}

	/**
	 * Allows sources to modify data returned by ES,
	 * before it goes to the client-side
	 *
	 * @param array &$result
	 * @param \Elastica\Result $resultObject
	 */
	public function format( &$result, $resultObject ) {
		// Base class format must work with original values
		// because it might be called multiple times
		$originalValues = $resultObject->getData();
		$result['type'] = $resultObject->getType();
		$result['score'] = $resultObject->getScore();

		// Experimental
		$user = $this->getContext()->getUser();
		if ( $user->isLoggedIn() ) {
			$resultRelevance = new \BS\ExtendedSearch\ResultRelevance( $user, $resultObject->getId() );
			$result['user_relevance'] = $resultRelevance->getValue();
		} else {
			$result['user_relevance'] = 0;
		}
		// End Experimental

		$type = $result['type'];
		$result['typetext'] = $this->getTypeText( $type );

		if ( $this->isFeatured( $result ) ) {
			$result['featured'] = 1;
		}

		if ( !isset( $originalValues['ctime'] ) || !isset( $originalValues['mtime'] ) ) {
			// Not all types have these
			return;
		}
		$result['ctime'] = $this->getContext()->getLanguage()->date( $originalValues['ctime'] );
		$result['mtime'] = $this->getContext()->getLanguage()->date( $originalValues['mtime'] );
	}

	/**
	 * Allows sources to modify results of autocomplete query
	 *
	 * @param array &$results
	 * @param array $searchData
	 */
	public function formatAutocompleteResults( &$results, $searchData ) {
		foreach ( $results as &$result ) {
			// For some reason _keys are not transmitted to client
			$result['id'] = $result['_id'];
			if ( !isset( $result['mtime'] ) || $result['rank'] !== 'top' ) {
				continue;
			}
			$result['modified_time'] = $this->getContext()->getLanguage()->timeanddate( $result['mtime'] );
			unset( $result['mtime'] );
		}
	}

	/**
	 *
	 * @param string $type
	 * @return string
	 */
	protected function getTypeText( $type ) {
		$typeText = $type;
		if ( wfMessage( "bs-extendedsearch-source-type-$type-label" )->exists() ) {
			$typeText = wfMessage( "bs-extendedsearch-source-type-$type-label" )->plain();
		}

		return $typeText;
	}

	/**
	 * Allows sources to change ranking of the autocomplete query results
	 * Exact matches are TOP, matches containing search term are NORMAL,
	 * and matches not containing search term (fuzzy) are SECONDARY
	 *
	 * Ranking controls where result will be shown( which part of AC popup )
	 *
	 * @param array &$results
	 * @param array $searchData
	 */
	public function rankAutocompleteResults( &$results, $searchData ) {
		$top = $this->getACHighestScored( $results );
		foreach ( $results as &$result ) {
			if ( $result['is_ranked'] == true ) {
				return;
			}

			$lcBasename = mb_strtolower( $result['basename'] );
			$lcSearchTerm = mb_strtolower( $searchData['value'] );
			if ( strpos( $lcBasename, $lcSearchTerm ) === 0 && $top['_id'] === $result['_id'] ) {
				$result['rank'] = self::AC_RANK_TOP;
			} elseif ( $this->matchTokenized( $lcBasename, $lcSearchTerm ) ) {
				$result['rank'] = self::AC_RANK_NORMAL;
			} else {
				$result['rank'] = self::AC_RANK_SECONDARY;
			}

			$result['is_ranked'] = true;
		}
	}

	/**
	 * Splits needle up in tokens and matches each token
	 * with the result title. If one token does not match
	 * whole matching is failed
	 *
	 * @param string $haystack
	 * @param string $needle
	 * @return bool
	 * @throws MWException
	 * @throws ConfigException
	 */
	protected function matchTokenized( $haystack, $needle ) {
		$separated = Wildcarder::factory( $needle )->getSeparated( [ '\s' ] );
		foreach ( $separated as $bit ) {
			if ( strpos( $haystack, $bit ) === false ) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Basic implementation. Checks if searched term
	 * matches result exactly
	 *
	 * @param array $result
	 * @return bool
	 */
	protected function isFeatured( $result ) {
		if ( $this->lookup == null ) {
			return false;
		}

		$queryString = $this->lookup->getQueryString();
		if ( isset( $queryString['query'] ) == false ) {
			return false;
		}

		$term = $queryString['query'];
		if ( strtolower( $term ) == strtolower( $result['basename'] ) ) {
			return true;
		}
	}

	/**
	 * Allows sources to modify filterCfg if needed
	 *
	 * @param array &$aggs
	 * @param array &$filterCfg
	 * @param bool $fieldsWithANDEnabled
	 */
	public function formatFilters( &$aggs, &$filterCfg, $fieldsWithANDEnabled = false ) {
	}

	/**
	 *
	 * @param array $results
	 * @return array|false
	 */
	protected function getACHighestScored( $results ) {
		$highest = false;
		foreach ( $results as $result ) {
			if ( !$highest || ( $result['score'] > $highest['score'] ) ) {
				$highest = $result;
			}
		}

		return $highest;
	}
}
